"""Functions for discretizing state-space systems."""

import numpy as np
from scipy.linalg import expm


def discretize_a(A, dt):
    """Discretizes the given continuous A matrix.

    Keyword arguments:
    A -- continuous system matrix
    dt -- discretization timestep

    Returns:
    discrete system matrix
    """
    # ϕ = eᴬᵀ
    return expm(A * dt)


def discretize_ab(A, B, dt):
    """Discretizes the given continuous A and B matrices.

    Keyword arguments:
    A -- continuous system matrix
    B -- continuous input matrix
    dt -- discretization timestep

    Returns:
    discrete system matrix, discrete input matrix
    """
    states = A.shape[0]
    inputs = B.shape[1]

    # M = [A  B]
    #     [0  0]
    M = expm(
        np.block([[A, B], [np.zeros((inputs, states)), np.zeros((inputs, inputs))]])
        * dt
    )

    # ϕ = eᴹᵀ = [A_d  B_d]
    #           [ 0    I ]
    return M[:states, :states], M[:states, states:]


def discretize_aq(A, Q, dt):
    """
    Discretizes the given continuous A and Q matrices.

    Keyword arguments:
    A -- Continuous system matrix.
    Q -- Continuous process noise covariance matrix.
    dt -- Discretization timestep.

    Returns:
    discrete system matrix, discrete process noise covariance matrix
    """
    states = A.shape[0]

    # Make continuous Q symmetric if it isn't already
    contQ = (Q + Q.T) / 2.0

    # M = [−A  Q ]
    #     [ 0  Aᵀ]
    M = np.block([[-A, contQ], [np.zeros((states, states)), A.T]])

    # ϕ = eᴹᵀ = [−A_d  A_d⁻¹Q_d]
    #           [ 0      A_dᵀ  ]
    phi = expm(M * dt)

    # ϕ₁₂ = A_d⁻¹Q_d
    phi12 = phi[:states, states:]

    # ϕ₂₂ = A_dᵀ
    phi22 = phi[states:, states:]

    discA = phi22.T
    discQ = discA @ phi12

    # Make discrete Q symmetric if it isn't already
    return discA, (discQ + discQ.T) / 2.0


def discretize_r(R, dt):
    """
    Returns a discretized version of the provided continuous measurement noise
    covariance matrix.

    Keyword arguments:
    R -- Continuous measurement noise covariance matrix.
    dt -- Discretization timestep.

    Returns:
    discrete measurement noise covariance matrix
    """
    # R_d = 1/T R
    return R / dt



"""Control system utility functions."""

import numpy as np


def make_cost_matrix(elems):
    """Creates a cost matrix from the given vector for use with LQR.

    The cost matrix is constructed using Bryson's rule. The inverse square
    of each element in the input is taken and placed on the cost matrix
    diagonal.

    Keyword arguments:
    elems -- a vector. For a Q matrix, its elements are the maximum allowed
             excursions of the states from the reference. For an R matrix,
             its elements are the maximum allowed excursions of the control
             inputs from no actuation.

    Returns:
    State excursion or control effort cost matrix
    """
    return np.diag([0.0 if elem == float("inf") else 1.0 / elem**2 for elem in elems])


def make_cov_matrix(elems):
    """Creates a covariance matrix from the given vector for use with Kalman
    filters.

    Each element is squared and placed on the covariance matrix diagonal.

    Keyword arguments:
    elems -- a vector. For a Q matrix, its elements are the standard
             deviations of each state from how the model behaves. For an R
             matrix, its elements are the standard deviations for each
             output measurement.

    Returns:
    Process noise or measurement noise covariance matrix
    """
    return np.diag(np.square(elems))


def ctrb(A, B):
    """Returns the controllability matrix for A and B

    Keyword arguments:
    A -- system matrix
    B -- input matrix
    """
    return np.hstack(
        [B] + [np.linalg.matrix_power(A, i) @ B for i in range(1, A.shape[0])]
    )


def obsv(A, C):
    """Returns the observability matrix for A and C

    Keyword arguments:
    A -- system matrix
    C -- output matrix
    """
    return np.vstack(
        [C] + [C @ np.linalg.matrix_power(A, i) for i in range(1, A.shape[0])]
    )



"""The WPILib state-space control library."""

from .ctrlutil import *
from .discretization import *
from .extended_kalman_filter import *
from .kalman_filter import *
from .linear_plant_inversion_feedforward import *
from .linear_quadratic_regulator import *
from .models import *
from .numerical_integration import *
from .numerical_jacobian import *
from .plotutil import *
from .profiles import *
